# @push.rocks/qenv
easy promised environments

## Install
To install `@push.rocks/qenv`, you need to have Node.js installed on your system. Once Node.js is installed, you can add `@push.rocks/qenv` to your project by running the following command in your project's root directory:

```bash
npm install @push.rocks/qenv --save
```

This command will add `@push.rocks/qenv` as a dependency to your project, and you will be ready to use it in your application.

## Usage
`@push.rocks/qenv` provides a convenient way to manage and access environment variables in your Node.js projects, especially when dealing with different environments like development, testing, and production. Its primary use is to load environment-specific variables in an easy and organized manner. Below is an extensive guide on how to use this module effectively in various scenarios, ensuring you can handle environment variables efficiently in your projects.

### Getting Started
First, ensure you have TypeScript configured in your project. `@push.rocks/qenv` is fully typed, providing excellent IntelliSense support when working in editors that support TypeScript, such as Visual Studio Code.

#### Importing Qenv
To get started, import the `Qenv` class from `@push.rocks/qenv`:

```typescript
import { Qenv } from '@push.rocks/qenv';
```

#### Basic Configuration
`@push.rocks/qenv` works with two main files: `qenv.yml` for specifying required environment variables, and `env.yml` for specifying values for these variables. These files should be placed in your project directory.

##### qenv.yml
This file specifies the environment variables that are required by your application. An example `qenv.yml` might look like this:

```yaml
required:
  - DB_HOST
  - DB_USER
  - DB_PASS
```

##### env.yml
This file contains the actual values for the environment variables in a development or testing environment. An example `env.yml` could be:

```yaml
DB_HOST: localhost
DB_USER: user
DB_PASS: pass
```

#### Instantiating Qenv
Create an instance of `Qenv` by providing paths to the directories containing the `qenv.yml` and `env.yml` files, respectively:

```typescript
const myQenv = new Qenv('./path/to/dir/with/qenv', './path/to/dir/with/env');
```

If the `env.yml` file is in the same directory as `qenv.yml`, you can omit the second argument:

```typescript
const myQenv = new Qenv('./path/to/dir/with/both');
```

#### Accessing Environment Variables
After instantiating `Qenv`, you can access the loaded environment variables directly from `process.env` in Node.js or through the `myQenv` instance for more complex scenarios like asynchronous variable resolution:

```typescript
// Accessing directly via process.env
console.log(process.env.DB_HOST); // 'localhost' in development environment

// Accessing via Qenv instance for more advanced scenarios
(async () => {
  const dbHost = await myQenv.getEnvVarOnDemand('DB_HOST');
  console.log(dbHost); // 'localhost'
})();
```

### Advanced Usage
#### Handling Missing Variables
By default, `Qenv` will throw an error and exit if any of the required environment variables specified in `qenv.yml` are missing. You can disable this behavior by passing `false` as the third argument to the constructor, which allows your application to handle missing variables gracefully:

```typescript
const myQenv = new Qenv('./path/to/dir/with/qenv', './path/to/dir/with/env', false);
```

#### Dynamic Environment Variables
For dynamic or computed environment variables, you can define functions that resolve these variables asynchronously. This is particularly useful for variables that require fetching from an external source:

```typescript
// Define a function to fetch a variable
const fetchDbHost = async () => {
  // Logic to fetch DB_HOST from an external service
  return 'dynamic.host';
};

// Use the function with getEnvVarOnDemand
(async () => {
  const dbHost = await myQenv.getEnvVarOnDemand(fetchDbHost);
  console.log(dbHost); // 'dynamic.host'
})();
```

#### Reading Variables from Docker Secrets or Other Sources
Internally, `@push.rocks/qenv` supports reading from Docker secrets, providing flexibility for applications deployed in Docker environments. The module attempts to read each required variable from the process environment, a provided `env.yml` file, Docker secrets, or any custom source you integrate.

### Conclusion
`@push.rocks/qenv` simplifies handling environment variables across different environments, making your application's configuration more manageable and secure. By separating variable definitions from their values and providing support for dynamic resolution, `@push.rocks/qenv` offers a robust solution for managing configuration in Node.js projects. Whether you're working in a local development environment, CI/CD pipelines, or production, `@push.rocks/qenv` ensures that you have the correct configuration for the task at hand.

Note: Due to the complexity and depth of `@push.rocks/qenv`, this documentation aims to cover general and advanced usage comprehensively. Please refer to the module's official documentation and typed definitions for further details on specific features or configuration options.

## License and Legal Information

This repository contains open-source code that is licensed under the MIT License. A copy of the MIT License can be found in the [license](license) file within this repository. 

**Please note:** The MIT License does not grant permission to use the trade names, trademarks, service marks, or product names of the project, except as required for reasonable and customary use in describing the origin of the work and reproducing the content of the NOTICE file.

### Trademarks

This project is owned and maintained by Task Venture Capital GmbH. The names and logos associated with Task Venture Capital GmbH and any related products or services are trademarks of Task Venture Capital GmbH and are not included within the scope of the MIT license granted herein. Use of these trademarks must comply with Task Venture Capital GmbH's Trademark Guidelines, and any usage must be approved in writing by Task Venture Capital GmbH.

### Company Information

Task Venture Capital GmbH  
Registered at District court Bremen HRB 35230 HB, Germany

For any legal inquiries or if you require further information, please contact us via email at hello@task.vc.

By using this repository, you acknowledge that you have read this section, agree to comply with its terms, and understand that the licensing of the code does not imply endorsement by Task Venture Capital GmbH of any derivative works.
